unit ParserMemoryStream;

{
    TParserMemoryStream v0.9
    A parser stream which stores data completely in memory. Faster than
    TParserFileStream for files, but loads the entire file into memory.
    
    Copyright (C) 2000 Ciaran McCreesh

    This library is available as part of the hktreeview distribution from
    <http://hktreeview.sourceforge.net/> or from the author via email
    <keesh@users.sourceforge.net>. Currently there is no standalone version
    as such, but if you're prepared to wait there will probably be one
    in the future. Otherwise, tweak away :)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ParserStream;

type
  TParserMemoryStream = class(TParserStream)
  private
    fStr: TMemoryStream;
  protected
    { Protected declarations }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function EOD: boolean; override;
    function Position: integer; override;
    function Seek(Offset: longint; Origin: word): integer; override;
    function CurrentChar: char; override;
    function ScanChar(const c: char): boolean; override;
    function ScanStr(const s: string): boolean; override;
    function ScanChars(const c: TChars): boolean; override;
    function Lookahead(const s: string): boolean; override;
    procedure LoadFromFile(const filename: string);
    procedure SaveToFile(const filename: string);
    procedure LoadFromString(const data: string);
    procedure Clear;
    function GetData(const offset, len: integer): string; override;
  published
    { Published declarations }
  end;

procedure Register;

implementation
{$R TPARSERMEMORYSTREAM.DCR}

procedure Register;
begin
  RegisterComponents('Parsers', [TParserMemoryStream]);
end;

{ TParserMemoryStream }

procedure TParserMemoryStream.Clear;
begin
  self.fStr.Clear;
end;

constructor TParserMemoryStream.Create(AOwner: TComponent);
begin
  inherited;
  self.fStr := TMemoryStream.Create;     
end;

function TParserMemoryStream.CurrentChar: char;
begin
  result := char(pointer(integer(self.fStr.Memory) + self.fStr.Position)^);  
end;

destructor TParserMemoryStream.Destroy;
begin
  self.fStr.Free;
  inherited;
end;

function TParserMemoryStream.EOD: boolean;
begin
  result := self.fStr.Position >= self.fStr.Size;
end;

function TParserMemoryStream.GetData(const offset, len: integer): string;
begin
  SetLength(result, len);
  CopyMemory(@result[1], pointer(integer(self.fStr.Memory) + offset), len);
end;

procedure TParserMemoryStream.LoadFromFile(const filename: string);
begin
  self.fStr.LoadFromFile(filename);
  self.Seek(0, soFromBeginning);
end;

procedure TParserMemoryStream.LoadFromString(const data: string);
begin
  self.fStr.Clear;
  self.fStr.Write(data[1], length(data));
  self.Seek(0, soFromBeginning);
end;

function TParserMemoryStream.Lookahead(const s: string): boolean;
var
  lens: integer;
begin
  lens := length(s);
  result := (self.fStr.Position + lens < self.fStr.Size) and
    (CompareMem(@s[1], pointer(integer(self.fStr.Memory) +
    self.fStr.Position + 1), lens));
end;

function TParserMemoryStream.Position: integer;
begin
  result := self.fStr.Position;
end;

procedure TParserMemoryStream.SaveToFile(const filename: string);
begin
  self.fStr.SaveToFile(filename);
end;

function TParserMemoryStream.ScanChar(const c: char): boolean;
begin
  result := false;
  while (not (self.EOD or result)) do begin
    self.fStr.Position := succ(self.fStr.Position);
    result := self.CurrentChar = c;
  end;
end;

function TParserMemoryStream.ScanChars(const c: TChars): boolean;
begin
  result := false;
  while (not (self.EOD or result)) do begin
    self.fStr.Position := succ(self.fStr.Position);
    result := self.CurrentChar in c;
  end;
end;

function TParserMemoryStream.ScanStr(const s: string): boolean;
var
  lens: integer;
begin
  lens := length(s);
  result := lens = 0;
  while (not (self.EOD or result)) do begin
    if (self.fStr.Position + lens >= self.fStr.Size) then
      self.Seek(0, soFromEnd)
    else
      result := self.ScanChar(s[1]) and CompareMem(@s[1],
        pointer(self.fStr.Position + integer(self.fStr.Memory)), lens);
  end;
end;

function TParserMemoryStream.Seek(Offset: Integer; Origin: word): integer;
begin
  self.fStr.Seek(Offset, Origin);
  result := self.Position;
end;

end.
