unit SXMLParser;

{
    TSXMLParser v0.9
    A general purpose XML / SGML parser. No knowledge of trees, just scans
    linearly. Other classes allow trees for HTML and XML documents to be
    created -- this class is the generic parser. Various properties control
    handling of !DOCTYPE declarations, nonexpanded boolean attributes,
    unquoted attribute values and so on, allowing various degrees of
    strictness.
    
    Copyright (C) 2000 Ciaran McCreesh

    This library is available as part of the hktreeview distribution from
    <http://hktreeview.sourceforge.net/> or from the author via email
    <keesh@users.sourceforge.net>. Currently there is no standalone version
    as such, but if you're prepared to wait there will probably be one
    in the future. Otherwise, tweak away :)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  SParser, StringHash, ParserStream;

type
  EXMLError = class(Exception);
  EXMLErrorSyntax = class(Exception);

  TSXMLElementKind = (xeEmpty, xeStart, xeEnd, xeComment, xePI, xeCDATA,
    xeUnknown, xeDTD);
  TSXMLProcessDTD = (pdAlways, pdAttempt, pdLocalOnly, pdNever);
  TSXMLParser = class(TSParser)
  private
    fStrictParsing: boolean;
    fStripComments: boolean;
    fProcessDTD: TSXMLProcessDTD;
    fElementName: string;
    fElementKind: TSXMLElementKind;
    fFastAttrs: boolean;
    fAttrs: TStringHash;
    fTextStart, fTextEnd: integer;
    procedure aSyntax(const assertion: boolean; const msg: string);
    procedure fParseAttrs;
  protected
    { Protected declarations }
  public
    property ElementName: string read fElementName stored false;
    property ElementKind: TSXMLElementKind read fElementKind;
    property Attributes: TStringHash read fAttrs;
    function NextElement: boolean;
    function ElementText: string;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property StripComments: boolean read fStripComments write fStripComments;
    property StrictParsing: boolean read fStrictParsing write fStrictParsing;
    property ProcessDTD: TSXMLProcessDTD read fProcessDTD write fProcessDTD;
    property FastAttrs: boolean read fFastAttrs write fFastAttrs;
  end;

const
  cElementNameDelim: TChars = [#$20, #$0a, #$0d, #$09, '>', '/'];
  cWhitespace: TChars = [#$20, #$0a, #$0d, #$09];

procedure Register;

implementation

{$R TSXMLPARSER.DCR}

procedure Register;
begin
  RegisterComponents('Parsers', [TSXMLParser]);
end;

{ TSXMLParser }

procedure TSXMLParser.aSyntax(const assertion: boolean; const msg: string);
begin
  if (not assertion) then
    raise EXMLErrorSyntax.Create(msg);
end;

constructor TSXMLParser.Create(AOwner: TComponent);
begin
  inherited;
  self.fStrictParsing := false;
  self.fStripComments := false;
  self.fProcessDTD := pdAttempt;
  self.fFastAttrs := true;
  self.fAttrs := TStringHash.Create(self);
end;

destructor TSXMLParser.Destroy;
begin
  self.fAttrs.Free;
  inherited;
end;

function TSXMLParser.ElementText: string;
begin
  if (self.fTextEnd >= self.fTextStart) then begin
    result := self.Stream.GetData(
      self.fTextStart,
      self.fTextEnd - self.fTextStart + 1);
  end else
    result := '';
end;

procedure TSXMLParser.fParseAttrs;

  procedure fSkipWhitespace;
  begin
    while (self.Stream.CurrentChar in cWhitespace) do
      self.Stream.Seek(1, soFromCurrent);
  end;

var
  finished: boolean;
  cs, ce: integer;
  qc: char;
  key: string;
begin
  { Zero or more attributes to parse. Each starts with cAttrNameChar+, followed
    by whitespace*. If strict parsing, next char must be '=', then whitespace*,
    then quote, then anything*, then quote. Otherwise, if next char is not '='
    then it's a nonexpanded boolean attribute, if it is '=' then skip
    whitespace* and take attribute value with quotes if included. }
  finished := false;
  while (not finished) do begin
    fSkipWhitespace;
    finished := self.Stream.CurrentChar in ['?', '/', '>'];
    if (not finished) then begin
      { get attribute name }
      cs := self.Position;
      aSyntax(self.Stream.ScanChars(cWhitespace + ['=']), 'Invalid Attribute Name');
      ce := self.Position - 1;
      key := self.Stream.GetData(cs, ce - cs + 1);
      fSkipWhitespace;
      aSyntax(self.Stream.CurrentChar = '=', 'Expected ''=''');
      self.Stream.Seek(1, soFromCurrent);
      fSkipWhitespace;
      qc := self.Stream.CurrentChar;
      if (qc in ['''', '"']) then begin
        cs := self.Position;
        aSyntax(self.Stream.ScanChar(qc), 'Expected ''' + qc + '''');
        ce := self.Position;
      end else begin
        cs := self.Position;
        aSyntax(self.Stream.ScanChars(cWhitespace + ['/', '>']),
          'Expected Attribute Value');
        self.Seek(-1, soFromCurrent);
        ce := self.Position;
      end;
      self.fAttrs[key] := self.Stream.GetData(cs, ce - cs + 1);
      self.Stream.Seek(1, soFromCurrent);
    end;
  end;
  if (self.Stream.CurrentChar = '/') then begin
    self.fElementKind := xeEmpty;
    self.Stream.Seek(1, soFromCurrent);
  end;
end;

function TSXMLParser.NextElement: boolean;
var
  cs, ce: integer;
begin
  { Normally can only get NextElement if not EOD and can find '<'. However,
    because ScanChar always skips one, an exception has to be added to allow
    for cases where there are two elements right next to each other. }
  result := (not self.EOD);
  self.fTextStart := self.Position;
  if (result) then
    result := (self.Stream.CurrentChar = '<') or self.Stream.ScanChar('<');

  self.fTextEnd := self.Position - 1;
  if (result) then begin
    { Clear any existing attributes }
    self.fAttrs.Clear;
    { Now at start of element. Element name is next character onwards. }
    self.Seek(1, soFromCurrent);
    cs := self.Position; { remember position of start of name }
    case (self.Stream.CurrentChar) of
      '!': begin { could be a comment, cdata or DTD-related }
        { lookahead for '--' (comment), '[CDATA[' (cdata) or something else }
        if (self.Stream.Lookahead('--')) then begin
          self.Stream.Seek(3, soFromCurrent);
          cs := self.Position; { move start of element name to after '--' }
          self.fElementKind := xeComment;
          aSyntax(self.Stream.ScanStr('-->'), 'Expected ''-->''');
          ce := self.Position - 1; { end of 'name' is previous char }
          self.Seek(3, soFromCurrent);
        end else if (self.Stream.Lookahead('[CDATA[')) then begin
          self.Stream.Seek(7, soFromCurrent);
          cs := self.Position; { move start of element name to after '[CDATA[' }
          self.fElementKind := xeCDATA;
          aSyntax(self.Stream.ScanStr(']]>'), 'Expected '']]>''');
          ce := self.Position - 1; { end of 'name' is previous char }
          self.Seek(3, soFromCurrent);
        end else begin
          self.fElementKind := xeDTD;
          aSyntax(self.Stream.ScanStr('>'), 'Expected ''>''');
          ce := self.Position - 1;
          self.Seek(1, soFromCurrent);
        end;
      end;
      '?': begin { Processing instruction or inline script stuff }
        { lookahead for 'xml ', otherwise just scan to next '?>' }
        if (self.Stream.Lookahead('xml ')) then begin
          self.fElementKind := xePI;
          ce := cs + 3; { gives '?xml' }
          self.Stream.Seek(4, soFromCurrent);
          self.fParseAttrs;
          self.Stream.Seek(2, soFromCurrent);
        end else begin
          self.fElementKind := xeUnknown;
          aSyntax(self.Stream.ScanStr('?>'), 'Expected ''?>''');
          ce := self.Position;
          self.Stream.Seek(2, soFromCurrent);
        end;
      end;
      '/': begin { End tag }
        { '/' then element name then whitespace* then '>' }
        self.fElementKind := xeEnd;
        inc(cs); { skip '/' }
        self.Stream.Seek(1, soFromCurrent);
        aSyntax(self.Stream.ScanChars(cElementNameDelim), 'Expected End of Element Name');
        aSyntax(self.Stream.CurrentChar = '>', 'Expected ''>''');
        ce := self.Position - 1;
        self.Stream.Seek(1, soFromCurrent);
      end;
      else begin { Start tag or empty element }
        self.fElementKind := xeStart; { fParseAttrs may change this to xeEmpty }
        aSyntax(self.Stream.ScanChars(cElementNameDelim), 'Expected End of Element Name');
        ce := self.Position - 1;
        self.fParseAttrs;
        self.Stream.Seek(1, soFromCurrent);
      end;
    end;
    self.fElementName := self.Stream.GetData(cs, ce - cs + 1);
  end;
end;

end.

